#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>
#include <assert.h>
#include <gen.h>
#include "utils.h"
#include "regex.h"
#include "hash.h"
#include "crc.h"

Version * cVersion = NULL;
BookmarkList * bmarklist = NULL;
char * bhome = NULL;
char * defaultBhome = BIBLEHOME;

#define MAXVERSIONS 4
Version versions[MAXVERSIONS];

#define NAME_HASH_NUM 1024
#define MARK_HASH_NUM 2048

void strlower(register char * str)
	{
	while (*str != EOS)
		{
		if (isupper(*str))
			*str = tolower(*str);
		str++;
		}
	}

void init()
	{
	struct stat sstat;
	int count;
	char bmpath[PATHSIZE+1];
	
	bhome = getenv("BIBLEHOME");
	if (bhome == NULL)
		bhome = defaultBhome;
	if (stat(bhome, &sstat) < 0 || !S_ISDIR(sstat.st_mode))
		{
		fprintf(stderr, "%s not found - set BIBLEHOME env variable\n", bhome);
		exit(1);
		}
	for (count = 0; count < MAXVERSIONS; count++)
		versions[count].title = NULL;
	cpystr(bmpath, bhome, "/config/", "bookmarks", NULL);
	if ((bmarklist = getBookmarks(bmpath)) == NULL)
		fprintf(stderr, "%s not Found\n", bmpath);
	}

#define VEC_INCREMENT 1024
#define MAX_LINE 1024

char ** getvector(char * filename, int * size)
	{
	char ** rtn = NULL;
	FILE * file;
	char buf[MAX_LINE];
	int num = 0;
	char * verName;
	int vsize = 0, popul = 0;
	
	rtn = (char **)malloc((vsize + 1) * sizeof (char *));
	/* in case the file is empty! */
	if ((file = fopen(filename, "r")) == NULL)
		{
#       ifdef DEB		
		errfmt("cannot open %s\n", filename);
#       endif		
		return NULL;
		}
	while (fgets(buf, sizeof(buf), file) != NULL)
		{
		buf[strlen(buf)-1] = '\0';
		if (popul == vsize)
			{
			vsize += VEC_INCREMENT;
			rtn = (char **)safe_realloc(rtn, (vsize + 1) * sizeof (char *));
			}
		/* use indexes with realloc because it's dangerous to fiddle
		   with pointers. realloc could move the memory! */
		rtn[popul] = strdup(buf);
		popul++;
		}
	rtn[popul] = NULL;
	fclose(file);
	if (size != NULL)
		*size = popul;
	return rtn;
	}

void freevector(char ** vec)
	{
	char ** ptr;
	
	ptr = vec;
	while (*ptr != NULL)
		free(*ptr++);
	free(vec);
	}

AbsVerse countVerses(char * book)
	{
	AbsVerse rtn = 0;
	AbsVerse dummy;
	register char * ptr;
	char last = '\0';
	Bool inVerse = NO;
	
	ptr = book;
	while (*ptr != '\0')
		{
		if (inVerse && *ptr == '\n' && last == '\n')
			{
			rtn++;
			inVerse = NO;
			}
		else if (!isspace(*ptr))
			inVerse = YES;
		last = *ptr++;
		}
	if (inVerse)
		rtn++;
#ifdef DEB	
	errfmt("NUM VERS: %d\n", rtn);
#endif	
	return rtn;
	}

static void setupVerses(char * book, Pos * verses, int numVerses)
	{
	int rtn = 1;
	register char * ptr;
	register int verseCount = 0;
	char last = '\0';
	Bool inVerse = NO;
	
	verses[verseCount++] = 0;
	ptr = book;
	while (*ptr != '\0' && verseCount < numVerses)
		{
		if (inVerse && *ptr == '\n' && last == '\n')
			{
			verses[verseCount++] = (ptr + 1) - book;
			inVerse = NO;
			}
		else if (!isspace(*ptr))
			inVerse = YES;
		last = *ptr++;
		}
/*
	if (inVerse)
		verses[verseCount++] = (ptr + 1) - book;
*/
	verses[verseCount] = -1;
	assert(verseCount == numVerses);
	}

void setupChapters(Book * book)
	{
	register AbsVerse count;
	ChapterNum chap;
	VerseNum ver;   

	parseVerseLine(book->data + book->verses[book->numVerses-1], &book->numChapters, &ver);
	book->chapters = (Chapter *)malloc(sizeof(Chapter) * (book->numChapters + 1));
	/* So we can index from 1! */
	/* Chapter 0 is sometimes used for introductions
	   so let it be valid!! */
	/* Maybe this text doesn't have chapters */
	if (book->numChapters != noChapterNum) 
		{
		for (chap = 0; chap <= book->numChapters; chap++)
			book->chapters[chap].begVerse = -1;
		for (count = 0; count < book->numVerses; count++)
			{
			parseVerseLine(book->data + book->verses[count], &chap, &ver);
			if (book->chapters[chap].begVerse == -1)
				book->chapters[chap].begVerse = count;
			book->chapters[chap].numVerses = ver;
			}
		}
	}

/* should the last \n be part of the length??? */
int verseLength(register char * verse)
	{
	register int rtn = 1;
#   ifdef DEB
	char temp[71];
	memmove(temp, verse, 70);
	temp[70] = EOS;
	errfmt("vl: %s\n", temp);
#   endif	
	while (!(*verse == '\0' || (*verse == '\n' && verse[1] == '\n')))
		{
		rtn++;
		verse++;
		}
	return rtn;
	}

void loadText(Book * book, FILE * file)
	{
	if (!book->textLoaded)
		{
		int offset;
#       ifdef DEB				
		errfmt("Load Text %s\n", book->title);
#       endif		
		book->data = (char *)malloc(book->size+2);
		if (fread(book->data, 1, book->size, file) != book->size)
			{
			eprint("Read Error: %s\n", errmesg(errno));
			exit(1);
			}
		pc2asc(book->data, book->size);
		offset = book->size;
		if (2 <= offset && 
			!(book->data[offset-1] == '\n' && book->data[offset-2] == '\n'))
			book->data[offset++] = '\n';
		book->data[offset] = '\0';
		book->textLoaded = YES;
		}
	}

FILE * openfile(char * filename)
	{
	FILE * rtn = NULL;
	char gzfilename[PATHSIZE+1];
	char gzcommand[PATHSIZE+1];
	struct stat sstat;
	
	cpystr(gzfilename, filename, ".gz", NULL);
	cpystr(gzcommand, "gunzip <", gzfilename, NULL);
	if ((rtn = fopen(filename, "r")) == NULL &&
		(0 <= stat(gzfilename, &sstat) && 
			(rtn = popen(gzcommand, "r")) == NULL))
		{
#       ifdef DEB			
		errfmt("cannot open file %s\n", filename);
#       endif			
		}
	return rtn;
	}

int filesize(FILE * file)
	{
	int rtn = 0;
	int res;
	char buf[4096];

	while (0 < (res = fread(buf, 1, sizeof(buf), file)))
		rtn += res;
	return rtn;
	}

Book * loadBookFile(Book * book, char * filename, Bool data)
	{
	struct stat sstat;
	FILE * file = NULL;
	
	if (!book->infoLoaded || (data && !book->textLoaded))
		{
		if ((file = openfile(filename)) == NULL)
			return NULL;
		else
			{
			if (!book->infoLoaded)
				{
#               ifdef DEB				
	        	errfmt("Loadf Info %s\n", book->title);
#               endif				
				book->size = filesize(file);
				fclose(file);
				if ((file = openfile(filename)) == NULL)
					return NULL;
				}
			loadText(book, file);
			assert(book->textLoaded);
			if (!book->infoLoaded)
				{
				book->numVerses = countVerses(book->data);
				book->verses = (Pos *)malloc((book->numVerses + 1) * sizeof(Pos));
				setupVerses(book->data, book->verses, book->numVerses);
				setupChapters(book);
				book->infoLoaded = YES;
				}
			fclose(file);
		}
	}
	return book;
}

void bookUnloadText(Book * book)
	{
	if (book->textLoaded)
		{
		free(book->data);
		book->textLoaded = NO;
		}
	}

int compareStrInt(register StringIntHash * item1, register StringIntHash * item2)
	{
	return strcasecmp(item1->str, item2->str);
	}

int hashStrInt(register StringIntHash * item)
{
	return hash_string_ignore_case(item->str);
}

int findBookIndex(Version * version, char * title)
	{
	StringIntHash * item, stmp;

	stmp.str = title;
	item = hash_get(&version->bookNameHash, &stmp);
	if (item == NULL)
		return -1;
	else
		return item->num;
	
#ifdef NEVER
	register Book * ptr;
	register Book * end;

	//TEMP
	end = &version->books[version->numBooks];
	for (ptr = &version->books[0]; ptr < end; ptr++)
		if (strcasecmp(ptr->title, title) == 0)
			break;
	if (end <= ptr)
		return -1;
	return ptr - version->books;
#endif // NEVER
	}

Book * findBook(Version * version, char * title)
	{
	int index;

	if ((index = findBookIndex(version, title)) < 0)
		return NULL;
	else
		return &version->books[index];
	}


Version * getFreeVersion()
	{
	int count;
	
	for (count = 0; count < MAXVERSIONS; count++)
		if (versions[count].title == NULL)
			return &versions[count];
	return NULL;
	}


void loadBook(Version * version, Book * book, Bool data)
	{
	if (!book->infoLoaded || (data && !book->textLoaded))
		{
		char pathname[PATHSIZE+1];
#ifdef NEVER
		char filename[NAMESIZE+1];
		register char * from = book->title;
		register char * to = filename;
		while (*from != EOS)
			if (*from == ' ')
				{
				*to++ = '_';
				from++;
				}
			else
				*to++ = tolower(*from++);
		*to++ = EOS;
#endif

		cpystr(pathname,
			   bhome,
			   "/versions/",
			   version->title,
			   "/",
			   book->title,
			   NULL);
		loadBookFile(book, pathname, data);
		assert(book->textLoaded);
		}
	}


void freeBook(Book * book)
	{
	if (book->infoLoaded)
		{
		bookUnloadText(book);
		free(book->verses);
		free(book->chapters);
		book->infoLoaded = NO;
		}
	}

#ifdef NEVER
int hashstr(char * str, int num)
	{
	return hash_string(str) % num;
	//TEMP
	u_long total, val;
	assert(0 < strlen(str));
	crc(str, strlen(str) + 1, &val, &total);
	return (int)(val % num);
	}
#endif // NEVER


Version * loadVersion(char * version)
	{
	Version * rtn;
	char filename[PATHSIZE+1];
	char ** titles;
	int count;
	BookNo numBooks;

	cpystr(filename,
		   bhome,
		   "/versions/",
		   version,
		   "/contents",
		   NULL);
	
	if ((titles = getvector(filename, &numBooks)) == NULL)
		return NULL;
	rtn = getFreeVersion();
	rtn->numBooks = numBooks;
	rtn->title = strdup(version);
	rtn->cBook = NULL;
	rtn->books = (Book *)malloc(sizeof(Book) * rtn->numBooks);
	rtn->bookNameHashItems = (StringIntHash *)malloc(sizeof(StringIntHash) * rtn->numBooks);
	hash_create(&rtn->bookNameHash, NAME_HASH_NUM,  (int (*)(void *, void *))compareStrInt, (int (*)(void *))hashStrInt, NULL);
	for (count = 0; count < rtn->numBooks; count++)
		{
		rtn->bookNameHashItems[count].str = strdup(titles[count]);	
		rtn->bookNameHashItems[count].num = count;
		hash_insert(&rtn->bookNameHash, &rtn->bookNameHashItems[count]);
		rtn->books[count].title = strdup(titles[count]);
		rtn->books[count].textLoaded = rtn->books[count].infoLoaded = NO;
#ifdef DEB 
		rtn->books[count].data = NULL;
		rtn->books[count].verses = NULL;
#endif		
		}
	freevector(titles);
	return rtn;
	}

Version * findVersion(char * title)
	{
	int count;
	
	for (count = 0; count < MAXVERSIONS; count++)
		if (versions[count].title != NULL && strcasecmp(versions[count].title, title) == EQUAL)
			return &versions[count];
	return loadVersion(title);
	}

/*
void freeVersion(Version * version)
{
		free(version->title);
		freevector(version->bookTitles);
		if (version->cBook != NULL)
			freeBook(version->cBook);
		free(version);
}
*/

void getVersionListFileName(char * filename)
	{
	cpystr(filename,
		   bhome,
		   "/versions/",
		   "/contents",
		   NULL);
	}

char ** getVersionListSVec(int * num)
	{
	char filename[PATHSIZE+1];
	
	getVersionListFileName(filename);
	return getvector(filename, num);
	}

void getBookContentsFileName(char * filename, char * version)
	{
	cpystr(filename,
		   bhome,
		   "/versions/",
		   version,
		   "/contents",
		   NULL);
	}

char ** getBookContentsSVec(char * ver, int * num)
	{
	char filename[PATHSIZE+1];
	
	getBookContentsFileName(filename, ver);
	return getvector(filename, num);
	}

Version * loadBookVersion(char * version, char * book)
	{
	Version * rtn;
	
	rtn = findVersion(version);
	rtn->cBook = findBook(rtn, book);
	loadBook(rtn, rtn->cBook, YES);
	return rtn;
	}

Bool loadBookAsCurrent(Version * ver, char * book)
	{
	Bool rtn = YES;
	if ((ver->cBook = findBook(ver, book)) == NULL)
		rtn = NO;
	else
		loadBook(ver, ver->cBook, YES);
	return rtn;
	}

void findDefaultBook(Version * ver)
	{
	if (ver->cBook == NULL)
		ver->cBook = findBook(ver, ver->books[0].title);
	}

void resetDefaultBook(Version * ver)
	{
	ver->cBook = findBook(ver, ver->books[0].title);
	}


int whatVerse(Book * book, int pos)
	{
	char * verse, *bpos;
	int count;
	
	bpos = book->data + pos;
	verse = book->data;
	for (count = book->numVerses - 1; 0 <= count; count--)
		{
		if (book->data + book->verses[count] <= bpos)
			return count;
		}
	return -1;
	}

char * allocVerseString(Book * book, int from, int to)
	{
	char * begin, *end, * rtn;
	int len;
	
	begin = book->data + book->verses[from];
	end = &(book->data + book->verses[to])[verseLength(book->data + book->verses[to])];
	len = end - begin;
	rtn = (char *)malloc(len + 1);
	memmove(rtn, begin, len);
	rtn[len] = EOS;
	return rtn;
	}

AbsVerse matchRange(Version * version, Book * book, int from, int to, regex_t * exp)
	{
	AbsVerse rtn = MAT_NOMATCH;
	char * buf;
	
	loadBook(version, book, YES);
	buf = allocVerseString(book, from, to);
	if (regexec(exp, buf, 0, NULL, 0) == 0)
		{
		rtn = from;
#ifdef DEB
		errfmt("Matched: %s absverse %d\n", buf, rtn);
#endif		
		}
	free(buf);
	return rtn;
	}

char * noMatch = "No Match Found";

AbsVerse searchNextBook(Version * version, Book * book, AbsVerse pos, Range * range, char * expression, ErrorBuf mesg, Bool back)
	{
	AbsVerse rtn = MAT_NOMATCH;
	regex_t cexp;
	int res;
	Verse verse;
	Triple rangeres;

	verse.title = book->title;
	verse.chap = noChapterNum;

/*
	assert(book != NULL && book->textLoaded);
*/
	if ((res = regcomp(&cexp, expression, 0)) != 0)
		{
		regerror(res, &cexp, mesg, ERRBUF_SIZE);
		rtn = MAT_ERROR;
		}
	else
		{
		int count, max;
		
		strcpy(mesg, noMatch);
		if (book->numVerses <= pos)
			return -1;
		max = book->numVerses;
		if (back)
		{
			for (count = pos; 0 <= count && rtn < 0; count--)
				{
#               ifdef DEB2
				errfmt("Trying %s verse %d\n", book->title, count);
#               endif				
				verse.verse = count;
				if ((rangeres = verseInRange(version, range, &verse, NO)) == YES)
					{
					if (0 <= (rtn = matchRange(version, book, count, count, &cexp)))
						break;
					}
				else if (rangeres == ERROR)
					{
					rtn = MAT_ERROR;
					strcpy(mesg, "Error in search Range");
					}
				}
			}
		else
			{
			for (count = pos; count < max && rtn < 0; count++)
				{
				verse.verse = count;
				if ((rangeres = verseInRange(version, range, &verse, NO)) == YES)
					{
					if (0 <= (rtn = matchRange(version, book, count, count, &cexp)))
						break;
					}
				else if (rangeres == ERROR)
				{
				rtn = MAT_ERROR;
				strcpy(mesg, "Error in search Range");
				}
				}
			}
		}
	return rtn;
	}

BookNo bookNum(Version * version, char * book)
	{
	BookNo rtn = -1, c;
	for (c = 0; c < version->numBooks; c++)
		{
		if (strcasecmp(version->books[c].title, book) == 0)
			{
			rtn = c;
			break;
			}
		}
	return rtn;
	}

Triple bookInRange(Version * version, Range * range, char * book)
	{
	Verse verse;
	Triple rangeres;

	verse.title = book;
	verse.chap = noChapterNum;
	verse.verse = noVerseNum;
	return verseInRange(version, range, &verse, YES);
	}

VerseList * newVerseList()
	{
	VerseList * rtn;

	rtn = (VerseList *)malloc(sizeof(VerseList));
	rtn->arraySize = 0;
	rtn->num = 0;
	}

void freeVerseList(VerseList * vl)
	{
	free(vl->list);
	free(vl);
	}

void addVerseList(VerseList * vl, Verse * verse)
	{
	if (vl->num == vl->arraySize)
		{
		vl->arraySize *= 2;
		if (vl->arraySize == 0)
			vl->arraySize = 8; /* arbitrary starting point */
		vl->list = (Verse *)safe_realloc(vl->list, sizeof(Verse) * vl->arraySize);
		}
	memcpy(&vl->list[vl->num], verse, sizeof(Verse));
	vl->num++;
	}

VerseList * searchVersionReturnAll(Version * version, Range * range, char * expression, ErrorBuf mesg)
{
/*
	VerseList * rtn;
	Bool nBook;
	Verse verse;
	AbsVerse pos = 0;
	

	con->verPosition = 0;
	resetDefaultBook(version);
	rtn = newVerseList();
	if ((pos = searchNextVersion(version, pos, range, search, errbuf, NO, &nBook)) != MAT_NOMATCH)
		{
		verse.title = version->cBook;
		absToChapVerse(version->cBook, pos, &verse.chap, &verse.verse);
		addVerseList(rtn, &verse);
		}
	return NO;
	}
	*/
}

AbsVerse searchNextVersion(Version * version, AbsVerse pos, Range * range, char * expression, ErrorBuf mesg, Bool back, Bool * newBook)
	{
	AbsVerse rtn = MAT_NOMATCH;
	Bool tryBook;
	
	assert(version != NULL);
	*newBook = NO;
	findDefaultBook(version);
	if (0 <= pos && bookInRange(version, range, version->cBook->title) == YES)
		{
		loadBookAsCurrent(version, version->cBook->title);
		rtn = searchNextBook(version, version->cBook, pos, range, expression, mesg, back);
		}
	if (rtn == MAT_NOMATCH)
		{
		BookNo num;
		Book * book;
		
		num = bookNum(version, version->cBook->title);
		if (back)
			num--;
		else
			num++;
		while (0 <= num && num < version->numBooks && rtn == MAT_NOMATCH)
			{
			tryBook = bookInRange(version, range, version->books[num].title) == YES;
			if (tryBook)
				loadBook(version, book = findBook(version, version->books[num].title), NO);
			if (back)
				{
				if (tryBook && (rtn = searchNextBook(version, book, book->numVerses - 1, range, expression, mesg, back)) < 0)
					bookUnloadText(book);
				num--;
				}
			else
				{
				if (tryBook && (rtn = searchNextBook(version, book, 0, range, expression, mesg, back)) < 0)
					bookUnloadText(book);
				num++;
				}
			}
		if (0 <= rtn)
			{
			*newBook = YES;
			assert(book != NULL);
			version->cBook = book;
			}
		}
	return rtn;
	}

Pos endOfVerse(Book * book, AbsVerse verse)
	{
	return book->verses[verse] +
		verseLength(book->data + book->verses[verse]);
	}

Bool isAtEndOfVerse(Book * book, AbsVerse verse, Pos pos)
	{
	return pos == endOfVerse(book, verse);
#ifdef NEVER
	return pos == book->verses[verse] +
		verseLength(book->verses[verse]) - book->data;
#endif // NEVER
	}

AbsVerse nextSearchVerse(Book * book, Pos pos)
	{
	AbsVerse rtn;
	
	rtn = whatVerse(book, pos);
	if (isAtEndOfVerse(book, rtn, pos))
		rtn++;
	if (book->numVerses <= rtn)
		rtn = -1;
	return rtn;
	}

AbsVerse prevSearchVerse(Book * book, Pos pos)
	{
	AbsVerse rtn;
	
	rtn = whatVerse(book, pos);
	if (isAtEndOfVerse(book, rtn, pos))
		rtn--;
	return rtn;
	}

void range_init(Range * range)
	{
	range->num = 0;
	}

RangeItem * rangeItem(Range * range, RangeNum num)
	{
	if (range->num <= num)
		{
		if (range->num == 0)
			range->items = (RangeItem *)malloc(1);
		range->num = num + 1;
		range->items = (RangeItem *)safe_realloc(range->items, sizeof(RangeItem) * range->num);
		}
	return &range->items[num];
	}

void verseInit(Verse * verse)
	{
	verse->title = NULL;
	verse->chap = noChapterNum;
	verse->verse = noVerseNum;
	}

void freeVerse(Verse * verse)
	{
	free(verse->title);
	}

void freeRangeItem(RangeItem * rangeItem)
	{
	freeVerse(&rangeItem->from);
	freeVerse(&rangeItem->to);
	}

void freeRange(Range * range)
	{
	RangeNum num;
	
	for (num = 0; num < range->num; num++)
		freeRangeItem(&range->items[num]);
	free(range);
	}

#define specialChar(c) ((c) == EOS || (c) == ':' || (c) == '-' || (c) == ',')

int parseV(char * verse, Verse * v)
	{
	char bbook[BOOKNAME_SIZE+1];
	char * book;
	register char *ptr;
	int num;
	
	ptr = verse;
	skipwhite(ptr);
	book = bbook;
	if (specialChar(*ptr))
		return -1;
	while (isdigit(*ptr))
		*book++ = *ptr++;
	while (!(specialChar(*ptr) || isdigit(*ptr) || isspace(*ptr)))
		*book++ = *ptr++;
	*book = EOS;
	skipwhite(ptr);
	if (*ptr == ':')
		{
		ptr = verse;
		skipwhite(ptr);
		}
	else
		{
#ifdef NEVER
		strlower(bbook);
#endif
		if (*bbook != EOS)
			v->title = strdup(bbook);
		}
	if (isdigit(*ptr))
		{
		num = atoi(ptr);
		v->chap = num;
		while (isdigit(*ptr))
			ptr++;
		skipwhite(ptr);
		if (*ptr == ':')
			{
			ptr++;
			skipwhite(ptr);
			num = atoi(ptr);
			v->verse = num;
			while (isdigit(*ptr))
				ptr++;
			}
		}
	return ptr - verse;
	}

Range * parseVerse(char * verse, int * eaten)
	{
	Range * rtn;
	RangeNum rnum = 0;
	RangeItem * ritem;
	register char *ptr;
	int res;
	
	rtn = (Range *)malloc(sizeof(Range));
	rtn->num = 0;
	ptr = verse;
	if (ptr != NULL)
		{
		skipwhite(ptr);
		while (*ptr != EOS)
			{
			ritem = rangeItem(rtn, rnum);
			verseInit(&ritem->from);
			verseInit(&ritem->to);
			if ((res = parseV(ptr, &ritem->from)) < 0)
				return NULL;
			ptr += res;
			skipwhite(ptr);
			if (*ptr == '-')
				{
				ptr++;
				ptr += parseV(ptr, &ritem->to);
				}
			if (*ptr == ',')
				ptr++;
			skipwhite(ptr);
			rnum++;
			}
		}
	*eaten = ptr - verse;
	return rtn;
	}

#ifdef NEVER
enum FAction { BEG, END, COPY, NONE, FAIL};

struct FixAction
	{
	FAction fchap;
	FAction fverse;
	FAction tchap;
	FAction tverse;
	};

struct FixAction actions[2][2][2][2] =
{
	{ /* from chap bad */
		{ /* from verse bad */
			{ /* to chap bad */
				{BEG, BEG, END, END}, /* to verse bad */
				{BEG, BEG, END, NONE}, /* to verse good */
			},
			{ /* to chap good */
				{BEG, BEG, NONE, END}, /* to verse bad */
				{BEG, BEG, NONE, NONE}, /* to verse good */
			}
		},
		{ /* from verse good */
			{ /* to chap bad */
				{BEG, NONE, COPY, END}, /* to verse bad */
				{BEG, NONE, COPY, NONE}, /* to verse good */
			},
			{ /* to chap good */
				NO, /* to verse bad */
				NO, /* to verse good */
			}
		}
	},
	{ /* from chap good */
		{ /* from verse bad */
			{ /* to chap bad */
				NO, /* to verse bad */
				NO, /* to verse good */
			},
			{ /* to chap good */
				NO, /* to verse bad */
				NO, /* to verse good */
			}
		},
		{ /* from verse good */
			{ /* to chap bad */
				NO, /* to verse bad */
				NO, /* to verse good */
			},
			{ /* to chap good */
				NO, /* to verse bad */
				{NONE, NONE, NONE, NONE}, /* to verse good */
			}
		}
	}
};
#endif				

void normaliseFromVerse(Verse * verse)
	{
	if (verse->chap == noChapterNum)
		verse->chap = 1;
	if (verse->verse == noVerseNum)
		verse->verse = 1;
	}

void normaliseToVerse(Version * version, RangeItem * ritem)
	{
	Book * book;
	
	if (ritem->to.chap == noChapterNum)
		{
		book = findBook(version, ritem->to.title);
		ritem->to.chap = book->numChapters;
		}
	if (ritem->to.verse == noVerseNum)
		{
		if (ritem->from.chap == noChapterNum)
			ritem->to.verse = book->chapters[book->numChapters].numVerses;
		}
	}
				
Bool normaliseRangeItem(Version * version, RangeItem * ritem)
	{
	if (ritem->from.title == NULL)
		return NO;
	if (ritem->to.title == NULL)
		ritem->to.title = strdup(ritem->from.title);
	else if (strcmp(ritem->to.title, "") == EQUAL)
		{
		free(ritem->to.title);
		ritem->to.title = strdup(version->books[version->numBooks-1].title);
		}
	normaliseFromVerse(&ritem->from);
	normaliseToVerse(version, ritem);
	}

Bool normaliseRange(Version * version, Range * range)
	{
	Bool rtn = YES;
	RangeNum rnum;
	RangeItem * ritem;
	
	for (rnum = 0; rnum < range->num; rnum++)
		{
		ritem = &range->items[rnum];
		if ((rtn = normaliseRangeItem(version, ritem)) == NO)
			break;
		}
	return rtn;
	}

#ifdef NEVER
int evaluateVerse(char * verse, char * book, ChapterNum * chap, VerseNum * verseNum)
	{
	char bbook[BOOKNAME_SIZE+1];
	ChapterNum bchap;
	VerseNum vnum;
	
	parseVerse(verse, book, chap, verseNum);
	}
#endif

AbsVerse gotoVerseString(Version * ver, char * verse, Bool * nBook)
	{
	char book[BOOKNAME_SIZE+1];
	ChapterNum chap;
	AbsVerse verseNum;
	int eaten;
	Range * range;
	Verse * v;

	if ((range = parseVerse(verse, &eaten)) == NULL)
		return noVerseNum;
	v = &range->items[0].from;
	return gotoVerse(ver, v->title, v->chap, v->verse, nBook);
	}

char * getAbsBeg(char * book, ChapterNum * chap, VerseNum * verseNum, int * recurs)
	{
	char * rtn;
	Bookmark * bm;

	rtn = book;
	bm = markFromList(bmarklist, book);
	if (bm != NULL)
		{
		book = bm->range->items[0].from.title;
		if (*chap == noChapterNum)
			*chap = bm->range->items[0].from.chap;
		if (*verseNum == noVerseNum)
			*verseNum = bm->range->items[0].from.verse;
		if ((*recurs)++ == 20) /* Just a big number */
			{
			fprintf(stderr, "%s hit bookmark recurs limit\n", book);
			return book;
			}
		rtn = getAbsBeg(book, chap, verseNum, recurs);
		if (*chap == noChapterNum)
			*chap = 1;
		if (*verseNum == noVerseNum)
			*verseNum = 1;
		return rtn;
		}
	return rtn;
	}

AbsVerse gotoVerse(Version * ver, char * bookName, ChapterNum chap, VerseNum verseNum, Bool * nBook)
	{
	AbsVerse rtn = noVerseNum;
	int recurs = 0;

	bookName = getAbsBeg(bookName, &chap, &verseNum, &recurs);
	if (*bookName == EOS)
		return noVerseNum;
	else if (ver->cBook == NULL || strcasecmp(bookName, ver->cBook->title) != 0)
		{
		Book * book;
		if (ver->cBook != NULL)  /* CB */
			bookUnloadText(ver->cBook);
		book = findBook(ver, bookName);
		if (book == NULL)
			return noVerseNum;
		loadBook(ver, ver->cBook = book, YES);
		*nBook = YES;
		}
	else
		*nBook = NO;
	if (ver->cBook == NULL) /* CB */
			return noVerseNum; /* CB */
	if (chap == noChapterNum)
		rtn = 0;   /* Verse #1 */
	else if (1 <= chap && chap <= ver->cBook->numChapters)
		{
		rtn = ver->cBook->chapters[chap].begVerse;
		if (verseNum != noVerseNum)
			rtn += (verseNum - 1);
		if (ver->cBook->numVerses <= rtn)
			rtn = noVerseNum;
		}
	return rtn;
	}

Bool absToChapVerse(Book * book, AbsVerse aver, ChapterNum * chap, VerseNum * vnum)
	{
	register ChapterNum count;
	Chapter * chapters = book->chapters;
	for (count = 0; count < book->numChapters; count++)
		if (chapters->begVerse <= aver && aver < chapters->begVerse + chapters->numVerses);
			{
			*chap = count;
			*vnum = aver - chapters->begVerse;
			return;
			}
	assert(NO);
	}


#ifdef NEVER
//TEMP
Pos verseToPos(Book * book, AbsVerse ver)
{
	return book->verses[ver] - book->data;
}
#endif // NEVER


void parseVerseLine(char * verse, ChapterNum * chap, VerseNum * verseNum)
	{
	char book[BOOKNAME_SIZE+1];
	Verse v;

	verseInit(&v);
	parseV(verse, &v);
	*chap = v.chap;
	*verseNum = v.verse;
	freeVerse(&v);
	/*	assert(isdigit(verse[res-1])); investigate!!! */
	}

int cmpBook(Version * version, char * b1, char * b2)
	{
	return findBookIndex(version, b1) - findBookIndex(version, b2);
	}

int cmpVerse(Version * version, Verse * v1, Verse * v2)
	{
	int rtn;
	if ((rtn = cmpBook(version, v1->title, v2->title)) != 0)
		return rtn;
	else if (v1->chap != noChapterNum && v2->chap != noChapterNum &&
			 (rtn = v1->chap - v2->chap) != 0)
		return rtn;
	else if (v1->verse != noVerseNum && v2->verse != noVerseNum && (rtn = v1->verse - v2->verse) != 0)
		return rtn;
	return 0;
	}

Bool nullVerse(Verse * verse)
	{
	return verse->title == NULL;
	}

Bool basicRange(Range * range)
	{
	return range->num == 1 && range->items[0].to.title == NULL;
	}

Triple verseInRange(Version * version, Range * range, Verse * verse, Bool onlyCompareBooks)
	{
	Triple rtn = NO;
	int eaten;
	RangeNum rnum;
	RangeItem * rangeItem;
	int res;

	if (range == NULL || range->num == 0)
		rtn = YES;
	else
		for (rnum = 0; rnum < range->num; rnum++)
			{
			Bookmark * markfrom = NULL, * markto = NULL;
			rangeItem = &range->items[rnum];
			assert(!nullVerse(&rangeItem->from));
			if (findBookIndex(version, rangeItem->from.title) < 0)
			{
				if (bmarklist == NULL)
					return ERROR;
				markfrom = markFromList(bmarklist, rangeItem->from.title);
				if (markfrom == NULL)
					return ERROR;
				}
			if (rangeItem->to.title != NULL &&
				findBookIndex(version, rangeItem->to.title) < 0)
				{
				if (bmarklist == NULL)
					return ERROR;
				markto = markFromList(bmarklist, rangeItem->to.title);
				if (markto == NULL)
					return ERROR;
				}
			if (!(markfrom == NULL && markto == NULL))
				{
				if (markfrom != NULL && rangeItem->to.title != NULL && !(basicRange(markfrom->range) && basicRange(markto->range)))
					return ERROR;
				if (rangeItem->to.title == NULL)
					{
					if (markfrom->range == NULL)
						return ERROR;
					rtn = verseInRange(version, markfrom->range, verse, onlyCompareBooks);
					}
				else
					{
					Range range2;
					RangeItem item2;
					
					assert(basicRange(markfrom->range));
					assert(markto == NULL || basicRange(markto->range));
					range2.items = &item2;
					range2.num = 1;
					if (markfrom == NULL)
						memmove(&item2.from, &rangeItem->from, sizeof(item2.from));
					else
						memmove(&item2.from, &markfrom->range->items[0].from, sizeof(item2.from));
					if (markto == NULL)
						memmove(&item2.to, &rangeItem->to, sizeof(item2.to));
					else
						memmove(&item2.to, &markto->range->items[0].from, sizeof(item2.to));
					rtn = verseInRange(version, &range2, verse, onlyCompareBooks);
					}
				if (rtn != NO)
					break;
				}
			else
				{
				if (onlyCompareBooks)
					{
					if (cmpBook(version, verse->title, rangeItem->from.title) == 0 ||
						(rangeItem->to.title != NULL &&
						 0 <= cmpBook(version, verse->title, rangeItem->from.title) &&
						 cmpBook(version, verse->title, rangeItem->to.title) <= 0))
						{
						rtn = YES;
						break;
						}
					}
				else
					{
					if (cmpVerse(version, verse, &rangeItem->from) == 0 ||
						(!nullVerse(&rangeItem->to) &&
						 0 <= cmpVerse(version, verse, &rangeItem->from) &&
						 cmpVerse(version, verse, &rangeItem->to) <= 0))
						{
						rtn = YES;
						break;
						}
					}
				}
			}
	return rtn;
	}

void catnotnull(char * buf, char * ps, int num)
	{
	char tbuf[256+1];
	if (0 <= num)
		{
		sprintf(tbuf, ps, num);
		strcat(buf, tbuf);
		}
	}

char * rangeToBuf(Range * range, char * buf)
	{
	int c;
	char tbuf[1024+1];
	
	buf[0] = EOS;
	for (c = 0; c < range->num; c++)
		{
		strcat(buf, range->items[c].from.title);
		catnotnull(buf, " %d", range->items[c].from.chap);
		catnotnull(buf, ":%d", range->items[c].from.verse);
		if (range->items[c].to.title != NULL)
			{
			strcat(buf, " - ");
			strcat(buf, range->items[c].to.title);
			catnotnull(buf, " %d", range->items[c].to.chap);
			catnotnull(buf, ":%d", range->items[c].to.verse);
			}
		if (c != range->num - 1)
			strcat(buf, ", ");
		}
	return buf;
	}

void bookmarksRebuildHash(BookmarkList * list)
	{
	register Bookmark * bm;
	register Bookmark * end;
	register StringIntHash * sh;
	BookmarkNum num = 0;

	hash_destroy(&list->bookmarkHash);
	hash_create(&list->bookmarkHash, MARK_HASH_NUM, (int (*)(void *, void *))compareStrInt, (int (*)(void *))hashStrInt, NULL);
	list->bookmarkHashItems = (StringIntHash *)safe_realloc(list->bookmarkHashItems, sizeof(StringIntHash) * list->num);
	for (bm = &list->bookmarks[0], end = &list->bookmarks[list->num], sh = &list->bookmarkHashItems[0];
		 bm < end; bm++, num++, sh++)
		{
		sh->str = strdup(bm->title);	
		sh->num = num;
		hash_insert(&list->bookmarkHash, sh);
		}
	}

BookmarkList * bookmark_create(BookmarkNum size)
	{
	BookmarkList * rtn;
	rtn = (BookmarkList *)malloc(sizeof(BookmarkList));
	rtn->bookmarks = (Bookmark *)malloc(sizeof(Bookmark) * size);
	hash_create(&rtn->bookmarkHash, MARK_HASH_NUM, (int (*)(void *, void *))compareStrInt, (int (*)(void *))hashStrInt, NULL);
	rtn->bookmarkHashItems = NULL;
	rtn->num = 0;
	bookmarksRebuildHash(rtn);
	return rtn;
	}

BookmarkList * getBookmarks(char * filename)
	{
	BookmarkList * rtn;
	FILE * file;
	char ** vec;
	char ** ptr;
	char * str, * pos;
	int size;
	int len, len2;
	int eaten;

	if ((vec = getvector(filename, &size)) == NULL)
		return NULL;
	rtn = bookmark_create(size);
	for (ptr = vec; *ptr != NULL; ptr++)
		{
		if (**ptr != EOS)
			{
			pos = strchr(*ptr, '=');
			if (pos != NULL)
				{
				len = pos - *ptr;
				str = rtn->bookmarks[rtn->num].title = (char *)malloc(len + 1);
				memmove(str, *ptr, len);
				str[len] = EOS;
				rtn->bookmarks[rtn->num].range = parseVerse(pos + 1, &eaten);
#ifdef NEVER
				len2 = strlen(*ptr) - len - 1;
				
				str = rtn->bookmarks[rtn->num].range = (char *)malloc(len2+1);
				memmove(str, *ptr + len + 1, len2);
				str[len2] = EOS;
#endif // NEVER
				rtn->num++;
				}
			}
		}
	freevector(vec);
	bookmarksRebuildHash(rtn);
	return rtn;
	}


BookmarkList * combineBookmarks(BookmarkList * l1, BookmarkList * l2)
	{
	BookmarkList * rtn;
	int count;
	
	rtn = (BookmarkList *)malloc(sizeof(BookmarkList));
	rtn->num = l1->num + l2->num;
	rtn->bookmarks = (Bookmark *)malloc(sizeof(Bookmark) * rtn->num);
	memmove(rtn->bookmarks, l1->bookmarks, sizeof(Bookmark) * l1->num);
	memmove(rtn->bookmarks + l1->num, l2->bookmarks, sizeof(Bookmark) * l2->num);
	bookmarksRebuildHash(rtn);
	return rtn;
	}

void bookmarkAdd(BookmarkList * bl, Bookmark * bm, BookmarkNum pos)
	{ 
	BookmarkNum move;
	if (pos < 0)
		pos = bl->num;
	move = bl->num - pos;
	bl->num++;
	bl->bookmarks = (Bookmark *)safe_realloc(bl->bookmarks, sizeof(Bookmark) * bl->num);
	memmove(bl->bookmarks + pos + 1, bl->bookmarks + pos, sizeof(Bookmark) * move); 
	memmove(bl->bookmarks + pos, bm, sizeof(Bookmark));
	bookmarksRebuildHash(bl);
	}

void bookmarkUpdate(BookmarkList * bl, Bookmark * bm, BookmarkNum num)
	{
	Bookmark * mk;

	mk = &bl->bookmarks[num];
	freeBookmark(mk);
	memmove(mk, bm, sizeof(Bookmark));
	bookmarksRebuildHash(bl);
	}

void bookmarkDelete(BookmarkList * bl, BookmarkNum num)
	{
	bl->num--;
	memmove(bl->bookmarks + num, bl->bookmarks + (num + 1), sizeof(Bookmark) * (bl->num - num));
	bl->bookmarks = safe_realloc(bl->bookmarks, sizeof(Bookmark) * bl->num);
	bookmarksRebuildHash(bl);
	}

void freeBookmark(Bookmark * bmark)
	{
	free(bmark->title);
	freeRange(bmark->range);
	}

void freeBookmarks(BookmarkList * list)
	{
	int count;
	
	for (count = 0; count < list->num; count++)
		freeBookmark(&list->bookmarks[count]);
	free(list->bookmarks);
	free(list);
	}

BookmarkList * reopenBookmark(char * filename, BookmarkList * old)
	{
	BookmarkList *list;
	if ((list = getBookmarks(filename)) != NULL)
		{
		if (old != NULL)
			freeBookmarks(old);
		old = list;
		}
	return old;
	}

Bookmark * markFromList(BookmarkList * list, char * name)
	{
	StringIntHash * item, stmp;

	stmp.str = name;
	item = hash_get(&list->bookmarkHash, &stmp);
	if (item == NULL)
		return NULL;
	else
		return &list->bookmarks[item->num];
#ifdef NEVER
	//TEMP
	Bookmark * rtn = NULL;
	register Bookmark * ptr;
	register Bookmark * end;
	int num;

	end = &list->bookmarks[list->num];
	if (list != NULL)
		for (ptr = &list->bookmarks[0]; ptr < end; ptr++)
			{
			if (strcasecmp(ptr->name, name) == EQUAL)
				{
				rtn = ptr;
				break;
				}
			}
	return rtn;
#endif // NEVER
}

BookmarkList * combinedLoadBookmark(char * filename, BookmarkList * old)
	{
	BookmarkList *list, * newlist;
	if ((list = getBookmarks(filename)) != NULL)
		{
		if (old == NULL)
			old = list;
		else
			{
			newlist = combineBookmarks(old, list);
			freeBookmarks(old);
			freeBookmarks(list);
			old = newlist;
			}
		}
	return old;
	}


